/*---------------------------------------------------------------------------*
 *                                   IT++			             *
 *---------------------------------------------------------------------------*
 * Copyright (c) 1995-2003 by Tony Ottosson, Thomas Eriksson, Pl Frenger,   *
 * Tobias Ringstrm, and Jonas Samuelsson.                                   *
 *                                                                           *
 * Permission to use, copy, modify, and distribute this software and its     *
 * documentation under the terms of the GNU General Public License is hereby *
 * granted. No representations are made about the suitability of this        *
 * software for any purpose. It is provided "as is" without expressed or     *
 * implied warranty. See the GNU General Public License for more details.    *
 *---------------------------------------------------------------------------*/

/*!
  \file
  \brief Vector copy functions for internal use
  \author Tony Ottosson

  1.8

  2003/06/23 12:01:22
*/

#include "itconfig.h"

#ifndef NO_CBLAS
#include "base/cblas.h"
#endif

using std::complex;

namespace itpp {

#ifndef DOXYGEN_SHOULD_SKIP_THIS

  template<class T> void copy_vector(const int n, const T *x, T *y);
  template<class T> void copy_vector(const int n, const T *x, const int incx, T *y, const int incy);
  template<class T> void swap_vector(const int n, T *x, T *y);
  template<class T> void swap_vector(const int n, T *x, const int incx, T *y, const int incy);

  /*
    Copy vector x to vector y. Both vectors are of size n
  */
#ifndef NO_CBLAS
  template<> inline
  void copy_vector(const int n, const double *x, double *y)
  {
    cblas_dcopy(n, x, 1, y, 1);
  }

  template<> inline
  void copy_vector(const int n, const std::complex<double> *x, std::complex<double> *y)
  {
    cblas_zcopy(n, x, 1, y, 1);
  }
#endif

  template<class T> inline
  void copy_vector(const int n, const T *x, T *y)
  {
    memcpy(y, x, (unsigned int)n*sizeof(T));
  }

  /*
    Copy vector x to vector y. Both vectors are of size n
    vector x elements are stored linearly with element increament incx
    vector y elements are stored linearly with element increament incx
  */
#ifndef NO_CBLAS
  template<> inline
  void copy_vector(const int n, const double *x, const int incx, double *y, const int incy)
  {
    cblas_dcopy(n, x, incx, y, incy);
  }

  template<> inline
  void copy_vector(const int n, const std::complex<double> *x, const int incx, std::complex<double> *y, const int incy)
  {
    cblas_zcopy(n, x, incx, y, incy);
  }
#endif

  template<class T> inline
  void copy_vector(const int n, const T *x, const int incx, T *y, const int incy)
  {
    for (int i=0;i<n; i++)
      y[i*incy] = x[i*incx];
  }

  /*
    Swap vector x to vector y. Both vectors are of size n
  */

#ifndef NO_CBLAS
  template<> inline
  void swap_vector(const int n, double *x, double *y)
  {
    cblas_dswap(n, x, 1, y, 1);
  }

  template<> inline
  void swap_vector(const int n, std::complex<double> *x, std::complex<double> *y)
  {
    cblas_zswap(n, x, 1, y, 1);
  }
#endif

  template<class T> inline
  void swap_vector(const int n, T *x, T *y)
  {
    for (int i=0; i<n; i++)
      swap(x[i], y[i]);
  }

  /*
    Swap vector x to vector y. Both vectors are of size n
    vector x elements are stored linearly with element increament incx
    vector y elements are stored linearly with element increament incx
  */
#ifndef NO_CBLAS
  template<> inline
  void swap_vector(const int n, double *x, const int incx, double *y, const int incy)
  {
    cblas_dswap(n, x, incx, y, incy);
  }

  template<> inline
  void swap_vector(const int n, std::complex<double> *x, const int incx, std::complex<double> *y, const int incy)
  {
    cblas_zswap(n, x, incx, y, incy);
  }
#endif

  template<class T> inline
  void swap_vector(const int n, T *x, const int incx, T *y, const int incy)
  {
    for (int i=0; i<n; i++)
      swap(x[i*incx], y[i*incy]);
  }

#endif /* DOXYGEN_SHOULD_SKIP_THIS */

} //namespace itpp
