/*---------------------------------------------------------------------------*
 *                                   IT++			             *
 *---------------------------------------------------------------------------*
 * Copyright (c) 1995-2003 by Tony Ottosson, Thomas Eriksson, Pl Frenger,   *
 * Tobias Ringstrm, and Jonas Samuelsson.                                   *
 *                                                                           *
 * Permission to use, copy, modify, and distribute this software and its     *
 * documentation under the terms of the GNU General Public License is hereby *
 * granted. No representations are made about the suitability of this        *
 * software for any purpose. It is provided "as is" without expressed or     *
 * implied warranty. See the GNU General Public License for more details.    *
 *---------------------------------------------------------------------------*/

/*! 
  \file 
  \brief Implementation of Communication Channel classes and functions.
  
  \author Pl frenger
  
  1.10
  
  2003/06/13 06:46:04
*/

#include "base/binary.h"
#include "base/matfunc.h"
#include "base/elmatfunc.h"
#include "base/specmat.h"
#include "comm/channel.h"
#include "base/filter.h"
#include "base/transforms.h"
#include "base/bessel.h"

namespace itpp {

  vec jake_filter(double NormFDopp, int order)
  {
    int i, L = (int)std::floor(double(order)/2);
    double t, x0;
    vec x_pos(L), x_neg(L), x(2*L+1), h(2*L+1);
    for (i=1; i<L+1; i++) {	
      t = double(i);
      x_pos(i-1) = besselj(0.25, 2*M_PI*NormFDopp*t) / std::sqrt(std::sqrt(t));
    }
    x0 = double(1.468813) * std::sqrt(std::sqrt(NormFDopp));
    x_neg = reverse(x_pos);
    x = concat( concat(x_neg, x0), x_pos);
    h = elem_mult(hamming(2*L+1), x);
    h /= norm(h);
    return h;
  }

  //------------------------------------------------------------------------------
  // Binary Symetric Channel, BSC.
  //------------------------------------------------------------------------------


  bvec BSC::operator()(const bvec &input)
  {
    int i, length = input.length();
    bvec output(length);
	
    for (i=0; i<length; i++) {
      if (u() <= p) {
	output(i) = input(i) + bin(1);
      } else {
	output(i) = input(i);
      }
    }
    return output;
  }

  //------------------------------------------------------------------------------
  // Channel
  //------------------------------------------------------------------------------

  //------------------------------------------------------------------------------
  // AWGNChannel
  //------------------------------------------------------------------------------

  cvec AWGN_Channel::operator()(const cvec &input)
  {
    cvec output = input + sigma*randn_c(input.size());
    return output;
  }

  vec AWGN_Channel::operator()(const vec &input)
  {
    vec output = input + sigma*randn(input.size());
    return output;
  }

  //------------------------------------------------------------------------------
  // Ricean Channel, Perfect Interleaving, RiceChannel
  //------------------------------------------------------------------------------

  cvec Rice_Channel::operator()(const cvec &input, cvec &chan)
  {
    double GG = std::pow(10.0,gamma/10.0);
    double Diffuse = std::sqrt( 1.0 / (1.0 + GG) );
    double Direct  = Diffuse * std::sqrt(GG);
    int i, length = input.length();
    complex<double> noise;
    cvec output(length);
    chan.set_size(length, false);
    
    for (i=0; i<length; i++) {
      noise = complex<double>(n()/M_SQRT2,n()/M_SQRT2);
      chan[i] = Direct + Diffuse * complex<double>(g(),g());
      output[i] = input[i] * chan[i] + noise;
    }
    return output;
  }

  cvec Rice_Channel::operator()(const cvec &input)
  {
    cvec dummy(input.length());
    cvec output = Rice_Channel::operator()(input,dummy);
    return(output);
  }

  vec Rice_Channel::operator()(const vec &input, vec &chan)
  {
    double GG = std::pow(10.0,gamma/10.0);
    double Diffuse = std::sqrt( 1.0 / (1.0 + GG) );
    double Direct = Diffuse * std::sqrt(GG);
    int i, length = input.length();
    double noise;
    vec output(length);
    chan.set_size(length, false);
	
    for (i=0; i<length; i++) {
      noise = n();
      chan[i] = abs( Direct + Diffuse * complex<double>(g(),g()));
      output[i] = input[i] * chan[i] + noise;
    }
    return (output);
  }

  vec Rice_Channel::operator()(const vec &input)
  {
    vec dummy(input.length());
    vec output = Rice_Channel::operator()(input,dummy);
    return(output);
  }

  //------------------------------------------------------------------------------
  // One Tap Rayleigh Fading Channel, Rayleigh_Channel
  //------------------------------------------------------------------------------

  Rayleigh_Channel::Rayleigh_Channel(double innoisevar, double innormdopp, int  injfiltord, int  inupsamp) : Channel(innoisevar)
  {
    normdopp = innormdopp;
    jfiltord = injfiltord; 
    upsamp   = inupsamp;
  }

  cvec Rayleigh_Channel::operator()(const cvec &input, cvec &chan)
  {
    int length = input.length();
    double temp;
    chan.set_size(length, false);
	
    if (normdopp==0) {
      //Perfect Interleaving:
      chan = randn_c(length);
      return ( elem_mult(input,chan) + sigma*randn_c(length) );	
    } else if (jfiltord>0) {
      //Use Jake-filter generation of fading process with linear interpolation:
      cvec jakeinput = randn_c((int)std::ceil(double(length)/upsamp)+jfiltord+2);
      vec jfilttaps = jake_filter(normdopp*upsamp,jfiltord);
      cvec jakeoutput = filter(to_cvec(jfilttaps),jakeinput);
      int i, u;
      complex<double> delta;	
      if (upsamp == 0) {
	chan = jakeoutput.mid(jfiltord+1,length);
      } else {
	for(i=0; i<std::floor(double(length)/upsamp); i++) {
	  delta = ( jakeoutput(i+jfiltord+2) - jakeoutput(i+jfiltord+1) ) / (double)upsamp;
	  for(u=0; u<upsamp; u++) {
	    chan(i*upsamp+u) = jakeoutput(i+jfiltord+1)+complex<double>(u,0.0)*delta;
	  }
	}
      }

      temp=std::sqrt(variance(chan));
      chan /= temp ;
      chan = elem_mult(input,chan);
      temp=std::sqrt(awgnvar);
      chan += temp*randn_c(length);
      return chan;
    } else {
      //Default. Use ifft generation of the fading process:	
      int Nfft = (int)round(std::pow(2,std::ceil(itpp::logb(2.0,double(length)))));
		
      //vec F = linspace(0,0.5,Nfft/2) ^ reverse(-linspace(0,0.5,Nfft/2));
      vec Fpos = linspace(0,0.5,Nfft/2+1); //Bugfix 2000-05-30 PF
      vec F = concat(Fpos, reverse(-Fpos(1,Nfft/2-1))); //Bugfix 2000-05-30 PF
		
		
      vec S = zeros(Nfft);
      double df = 1.0/double(Nfft);
      int noisesamp = (int)std::ceil(normdopp/df);
		
      for (int i=0; i<F.length(); i++) {
	if (std::abs(F(i))<normdopp)
	  S(i) = std::sqrt(1.5 / ( M_PI*normdopp*std::sqrt(1-std::pow(F(i)/normdopp,2))));
	else if (std::abs(F(i))==normdopp)
	  S(i) = 1000000;
      }
		
      S /= norm(S);
      S *= Nfft;
		
      cvec x = ifft( elem_mult(to_cvec(S), concat(randn_c(noisesamp), zeros_c(Nfft-2*noisesamp), randn_c(noisesamp))) ); 
      chan = x.mid(0,length);
		
      return ( elem_mult(input,chan) + sigma*randn_c(length) );
    }
  }

  cvec Rayleigh_Channel::operator()(const cvec &input) 
  {
    cvec dummy(input.length());
    return( Rayleigh_Channel::operator()(input, dummy) );
  }

  vec Rayleigh_Channel::operator()(const vec &input, vec &chan)
  {
    int length = input.length();
    chan.set_length(length, false);
    if (normdopp==0) {
      //Perfect Interleaving:
      chan = randray(length,1.0/std::sqrt(2.0));
      return ( elem_mult(input, chan) + sigma*randn(length) );
    } else if (jfiltord>0) {
      //Use Jake-filter generation of fading process with linear interpolation:
      cvec jakeinput  = randn_c((int)std::ceil(double(length)/upsamp)+jfiltord+2);
      vec jfilttaps  = jake_filter(normdopp*upsamp,jfiltord);
      vec jakeoutput = abs(filter(to_cvec(jfilttaps),jakeinput));
      int i, u;
      double delta;
      if (upsamp==0) {
	chan = jakeoutput.mid(jfiltord+1,length);
      } else {
	for(i=0; i<std::floor(double(length)/upsamp); i++) {
	  delta = ( jakeoutput(i+jfiltord+2) - jakeoutput(i+jfiltord+1) ) / upsamp;
	  for(u=0; u<upsamp; u++) {
	    chan(i*upsamp+u) = jakeoutput(i+jfiltord+1)+u*delta;
	  }
	}
      }
      return( elem_mult(input,chan) + sigma*randn(length) );
    } else { 
      //Default. Use ifft generation of the fading process:	
      int Nfft = (int)round(std::pow(2,std::ceil(logb(2.0,double(length)))));
		
      //vec F = linspace(0,0.5,Nfft/2) ^ reverse(-linspace(0,0.5,Nfft/2));
      vec Fpos = linspace(0,0.5,Nfft/2+1); //Bugfix 2000-05-30 PF
      vec F = concat(Fpos, reverse(-Fpos(1,Nfft/2-1))); //Bugfix 2000-05-30 PF
		
      vec S = zeros(Nfft);
      double df = 1.0/Nfft;
      int noisesamp = (int)std::ceil(normdopp/df);
      for (int i=0; i<F.length(); i++) {
	if (std::abs(F(i))<normdopp)
	  S(i) = std::sqrt(1.5 / ( M_PI*normdopp*std::sqrt(1-std::pow(F(i)/normdopp,2))));
	else if (std::abs(F(i))==normdopp)
	  S(i) = 1000000;
      }
      S /= norm(S);
      S *= Nfft;
      vec x = abs( ifft( elem_mult(to_cvec(S), concat(randn_c(noisesamp), zeros_c(Nfft-2*noisesamp), randn_c(noisesamp))) ) ); 
      chan = x.mid(0,length);
      return ( elem_mult(input,chan) + sigma*randn(length) );
    }
  }

  vec Rayleigh_Channel::operator()(const vec &input)
  {
    vec dummy(input.length());
    return( Rayleigh_Channel::operator()(input, dummy) );
  }


  void Rayleigh_Channel::create_channel(int no_samples, cvec &channel)
  {
    double	temp;
    channel.set_size(no_samples, false); 
	
    if (normdopp==0) {
      //Perfect Interleaving:
      channel = randn_c(no_samples);
    } else if (jfiltord>0) {
      //Use Jake-filter generation of fading process with linear interpolation:
      cvec jakeinput  = randn_c((int)std::ceil(double(no_samples)/upsamp)+jfiltord+2);
      vec jfilttaps  = jake_filter(normdopp*upsamp,jfiltord);
      cvec jakeoutput = filter(to_cvec(jfilttaps),jakeinput);
      int i, u;
      complex<double> delta;	
      if (upsamp==0) {
	channel = jakeoutput.mid(jfiltord+1, no_samples);
      } else {
	for(i=0; i<std::floor(double(no_samples)/upsamp); i++) {
	  delta = ( jakeoutput(i+jfiltord+2) - jakeoutput(i+jfiltord+1) ) / (double)upsamp;
	  for(u=0; u<upsamp; u++) {
	    channel(i*upsamp+u) = jakeoutput(i+jfiltord+1)+complex<double>(u,0.0)*delta;
	  }
	}
      }
		
      temp=std::sqrt(variance(channel));
      channel /= temp ;
    } else {
      //Default. Use ifft generation of the fading process:	
      int Nfft = (int)round(std::pow(2,std::ceil(logb(2.0,double(no_samples)))));
		
      //vec F = linspace(0,0.5,Nfft/2) ^ reverse(-linspace(0,0.5,Nfft/2));
      vec Fpos = linspace(0,0.5,Nfft/2+1); //Bugfix 2000-05-30 PF
      vec F = concat(Fpos, reverse(-Fpos(1,Nfft/2-1))); //Bugfix 2000-05-30 PF
      vec S = zeros(Nfft);
      double df = 1.0/double(Nfft);
      int noisesamp = (int)std::ceil(normdopp/df);
		
      for (int i=0; i<F.length(); i++) {
	if (std::abs(F(i))<normdopp)
	  S(i) = std::sqrt(1.5 / ( M_PI*normdopp*std::sqrt(1-std::pow(F(i)/normdopp,2))));
	else if (std::abs(F(i))==normdopp)
	  S(i) = 1000000;
      }
		
      S /= norm(S);
      S *= Nfft;
		
      cvec x = ifft( elem_mult(to_cvec(S), concat(randn_c(noisesamp), zeros_c(Nfft-2*noisesamp), randn_c(noisesamp))) ); 
      channel = x.mid(0, no_samples);
    }
  }


  cvec Rayleigh_Channel::filter(const cvec &channel, const cvec &input)
  {
    it_assert1(channel.size() == input.size(), "Rayleigh_Channel::filter. Wrong size");
    return ( elem_mult(input,channel) );
  }


  //------------------------------------------------------------------------------
  // Multi Path Rayleigh Fading Channel, Multipath_Rayleigh_Channel
  //------------------------------------------------------------------------------

  Multipath_Rayleigh_Channel::Multipath_Rayleigh_Channel(double inawgnvar, double innormdopp, 
							 short innopaths, ivec indelayvect, vec inpowvecdb) : Channel(inawgnvar)
  {
    normdopp = innormdopp;
    nopaths  = innopaths;
    delvec   = indelayvect;
    powvecdb = inpowvecdb;
  }

  cvec Multipath_Rayleigh_Channel::operator()(const cvec &input, cmat &chan)
  {
    int i, length = input.length();
    chan.set_size(nopaths, length, false);
    //default: Delays are sample spaced "0:nopaths-1".
    if (delvec.length()==1 && delvec(0) == 0) {
      delvec.set_size(nopaths, false);
      for (i=0; i<nopaths; i++) {
	delvec(i) = i;
      }
    }
    //default: Uniform power distribution.
    if (powvecdb.length()==1 && powvecdb(0) == 0) {
      powvecdb.set_length(nopaths, false);
      powvecdb = zeros(nopaths);
    }
    vec	powvec = pow(10.0,powvecdb/static_cast<double>(10.0));
    powvec /= norm(powvec);
    int maxdelay = max(delvec);
    cvec out = zeros_c(length+maxdelay);
    int	Nfft = (int)std::pow(2,std::ceil(logb(2.0,double(length))));
	
    //vec F = linspace(0,0.5,Nfft/2) ^ reverse(-linspace(0,0.5,Nfft/2));
    vec Fpos = linspace(0,0.5,Nfft/2+1); //Bugfix 2000-05-30 PF
    vec F = concat(Fpos, reverse(-Fpos(1,Nfft/2-1))); //Bugfix 2000-05-30 PF
	
    vec	S = zeros(Nfft);
    double df = 1.0/Nfft;
    int	noisesamp = (int)std::ceil(normdopp/df);
    cvec x(Nfft), y(length);
	
    for (i=0; i<F.length(); i++) {
      if (std::abs(F(i))<normdopp)
	S(i) = std::sqrt(1.5 / ( M_PI*normdopp*std::sqrt(1-std::pow(F(i)/normdopp,2))));
      else if (std::abs(F(i))==normdopp)
	S(i) = 1000000;
    }
	
    S /= norm(S,2);
    S *= Nfft;
	
    for (i=0; i<nopaths; i++) {
      x = ifft( elem_mult(to_cvec(S), concat(randn_c(noisesamp), zeros_c(Nfft-2*noisesamp), randn_c(noisesamp)) ) ); 
      y = x.mid(0,length);
      chan.set_row( i, y*complex<double>(powvec(i),0) );
      out += concat( zeros_c(delvec(i)), elem_mult(input,chan.get_row(i)), zeros_c(maxdelay-delvec(i)) );
    }
	
    return ( out + sigma*randn_c(length+maxdelay) );
  }

  cvec Multipath_Rayleigh_Channel::operator()(const cvec &input) {
    cmat dummy(input.length(),nopaths);
    return( Multipath_Rayleigh_Channel::operator()(input, dummy) );
  }


  void Multipath_Rayleigh_Channel::create_channel(int no_samples, cmat &channel)
  {
    int 	i;
    channel.set_size(nopaths, no_samples, false);
    //default: Delays are sample spaced "0:nopaths-1".
    if (delvec.length()==1 && delvec(0) == 0) {
      delvec.set_size(nopaths, false);
      for (i=0; i<nopaths; i++) {
	delvec(i) = i;
      }
    }
    //default: Uniform power distribution.
    if (powvecdb.length()==1 && powvecdb(0) == 0) {
      powvecdb.set_length(nopaths, false);
      powvecdb = zeros(nopaths);
    }
    vec	powvec = pow(10.0,powvecdb/static_cast<double>(10.0));
    powvec /= norm(powvec);
    int	Nfft = (int)std::pow(2,std::ceil(logb(2.0,double(no_samples))));
	
    // vec F = linspace(0,0.5,Nfft/2) ^ reverse(-linspace(0,0.5,Nfft/2));
    vec Fpos = linspace(0,0.5,Nfft/2+1); //Bugfix 2000-05-30 PF
    vec F = concat(Fpos, reverse(-Fpos(1,Nfft/2-1))); //Bugfix 2000-05-30 PF
	
    vec	S = zeros(Nfft);
    double df = 1.0/Nfft;
    int	noisesamp = (int)std::ceil(normdopp/df);
    cvec x(Nfft), y(no_samples);
	
    for (i=0; i<F.length(); i++) {
      if (std::abs(F(i))<normdopp)
	S(i) = std::sqrt(1.5 / ( M_PI*normdopp*std::sqrt(1-std::pow(F(i)/normdopp,2))));
      else if (std::abs(F(i))==normdopp)
	S(i) = 1000000;
    }
	
    S /= norm(S,2);
    S *= Nfft;
	
    for (i=0; i<nopaths; i++) {
      x = ifft( elem_mult(to_cvec(S), concat(randn_c(noisesamp), zeros_c(Nfft-2*noisesamp), randn_c(noisesamp)) ) ); 
      y = x.mid(0,no_samples);
      channel.set_row( i, y*complex<double>(powvec(i),0) );
    }
  }


  cvec Multipath_Rayleigh_Channel::filter(const cmat &channel, const cvec &input)
  {
    it_assert1(channel.cols() == input.size(), "Multipath_Rayleigh_Channel::filter. Wrong size");
    it_assert1(channel.rows() == nopaths, "Multipath_Rayleigh_Channel::filter. Wrong size");
    int i, length = input.length();
    int maxdelay = max(delvec);
    cvec out(length+maxdelay);
    out.zeros();
	
    for (i=0; i<nopaths; i++) {
      out += concat( zeros_c(delvec(i)), elem_mult(input,channel.get_row(i)), zeros_c(maxdelay-delvec(i)) );
    }
	
    return out;
  }

} //namespace itpp
