/*---------------------------------------------------------------------------*
 *                                   IT++			             *
 *---------------------------------------------------------------------------*
 * Copyright (c) 1995-2001 by Tony Ottosson, Thomas Eriksson, Pl Frenger,   *
 * Tobias Ringstrm, and Jonas Samuelsson.                                   *
 *                                                                           *
 * Permission to use, copy, modify, and distribute this software and its     *
 * documentation under the terms of the GNU General Public License is hereby *
 * granted. No representations are made about the suitability of this        *
 * software for any purpose. It is provided "as is" without expressed or     *
 * implied warranty. See the GNU General Public License for more details.    *
 *---------------------------------------------------------------------------*/

/*! 
  \file 
  \brief Implementation of audio device classes.
  \author Tobias Ringstrm

  1.4

  2003/05/22 08:55:20
*/

#include "srccode/audiodev.h"

// ---------------- Linux audio code ----------------------------------------
#ifdef linux

#include <linux/soundcard.h>
#include <iostream>
#include <cstdlib>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>

namespace itpp {

#define AUDIO_DEVICE "/dev/audio"

  inline static signed char clip_to_char(double x)
  {
    if (x < -127.0)
      return -127;
    else if (x > 127.0)
      return 127;
    else
      return (signed char)(x+0.5);
  }


  inline static short clip_to_short(double x)
  {
    if (x < -32767.0)
      return -32767;
    else if (x > 32767.0)
      return 32767;
    else
      return (short)(x+0.5);
  }


  struct Audio_Basic_Device::Device {
    int audioDev;
  };

  //
  // Audio_Basic_Device
  //
  Audio_Basic_Device::Audio_Basic_Device()
  {
    dev = new Device;
    isOpen = false;
  }


  Audio_Basic_Device::~Audio_Basic_Device()
  {
    close();
    delete dev;
  }


  bool Audio_Basic_Device::open(open_mode mode)
  {
    int flags = 0;
	
    switch (mode) {
    case mode_rd :
      flags = O_RDONLY;
      break;
    case mode_wr :
      flags = O_WRONLY;
      break;
    case mode_rdwr :
      flags = O_RDWR;
      break;
    }
	
    dev->audioDev = ::open(AUDIO_DEVICE, flags);
    isOpen = (dev->audioDev != -1);
	
    return isOpen;
  }


  void Audio_Basic_Device::close()
  {
    if (isOpen)
      ::close(dev->audioDev);
    isOpen = false;
  }


  int Audio_Basic_Device::read(signed char *p, int n)
  {
    return ::read(dev->audioDev, p, n * sizeof(signed char)) / sizeof(signed char);
  }


  int Audio_Basic_Device::read(short *p, int n)
  {
    return ::read(dev->audioDev, p, n * sizeof(short)) / sizeof(short);
  }


  int Audio_Basic_Device::write(const signed char *p, int n)
  {
    return ::write(dev->audioDev, p, n * sizeof(signed char)) / sizeof(signed char);
  }


  int Audio_Basic_Device::write(const short *p, int n)
  {
    return ::write(dev->audioDev, p, n * sizeof(short)) / sizeof(short);
  }


  //
  // Audio_Play_Device
  //
  Audio_Play_Device::Audio_Play_Device(double max_amplitude,
				       int rate, int bits, bool stereo)
  {
    play_max_ampl = max_amplitude;
    it_assert1(open(mode_wr), "Cannot open audio device!");
    
    set_play_rate(rate);
    set_play_bits(bits);
    enable_play_stereo(stereo);
  }

	
  Audio_Play_Device::Audio_Play_Device(double max_amplitude)
  {
    play_max_ampl = max_amplitude;
  }


  Audio_Play_Device::~Audio_Play_Device()
  {
    close();
  }


  int Audio_Play_Device::get_play_rate()
  {
    return play_rate;
  }


  bool Audio_Play_Device::set_play_rate(int rate)
  {
    ioctl(dev->audioDev, SNDCTL_DSP_SPEED, &rate);
    play_rate = rate;
    return true;
  }


  int Audio_Play_Device::get_play_bits()
  {
    return play_bits;
  }


  bool Audio_Play_Device::set_play_bits(int bits)
  {
    int fmt = bits==16 ? AFMT_S16_LE : AFMT_S8;

    if (bits!=8 && bits!=16)
      return false;
	
    ioctl(dev->audioDev, SNDCTL_DSP_SETFMT, &fmt);
    play_bits = bits;
	
    return true;
  }


  bool Audio_Play_Device::query_play_stereo()
  {
    return play_stereo;
  }


  bool Audio_Play_Device::enable_play_stereo(bool on)
  {
    ioctl(dev->audioDev, SNDCTL_DSP_STEREO, &on);
    play_stereo = on;
    return true;
  }


  int Audio_Play_Device::get_play_buffersize()
  {
    int size;
    ioctl(dev->audioDev, SNDCTL_DSP_GETBLKSIZE, &size);
    
    return size;
  }

  // TODO: This should probably be implemented
  bool Audio_Play_Device::set_play_buffersize(int size)
  {
    return false;
  }


  bool Audio_Play_Device::play(const vec &mono)
  {
    int i, n;

    if (play_stereo)
      return play(mono, mono);

    if (play_bits == 8) {
      signed char data[mono.size()];
      for (i=0; i<mono.size(); i++)
	data[i] = clip_to_char(mono(i)*127.0/play_max_ampl);
      n = write(data, mono.size());
    } else {
      short data[mono.size()];
      for (i=0; i<mono.size(); i++)
	data[i] = clip_to_short(mono(i)*32767.0/play_max_ampl);
      n = write(data, mono.size());
    }

    return n == mono.size();
  }


  bool Audio_Play_Device::play(const vec &left, const vec &right)
  {
    int i, n;

    it_assert1(left.size() == right.size(), "Left and right channel must have the same size in play()!");
    
    if (!play_stereo)
      return play(left + right / 2.0);

    if (play_bits == 8) {
      signed char data[2 * left.size()];
      for (i=0; i<left.size(); i++) {
	data[2*i] = clip_to_char(left(i)*127.0/play_max_ampl);
	data[2*i+1] = clip_to_char(right(i)*127.0/play_max_ampl);
      }
      n = write(data, 2 * left.size()) / 2;
    } else {
      short data[2 * left.size()];
      for (i=0; i<left.size(); i++) {
	data[2*i] = clip_to_short(left(i)*32767.0/play_max_ampl);
	data[2*i+1] = clip_to_short(right(i)*32767.0/play_max_ampl);
      }
      n = write(data, 2 * left.size()) / 2;
    }

    return n == left.size();
  }


  void Audio_Play_Device::stop_play()
  {
    it_error("Audio_Play_Device::stop_play() needs to be written!");
  }


  void Audio_Play_Device::flush()
  {
    it_error("Audio_Play_Device::flush() needs to be written!");
  }


  //
  // Audio_Rec_Device
  //
  Audio_Rec_Device::Audio_Rec_Device(double max_amplitude,
				     int rate, int bits, bool stereo)
  {
    rec_max_ampl = max_amplitude;
    
    it_assert1(open(mode_rd), "Cannot open audio device!");

    set_rec_rate(rate);
    set_rec_bits(bits);
    enable_rec_stereo(stereo);
  }


  Audio_Rec_Device::Audio_Rec_Device(double max_amplitude)
  {
    rec_max_ampl = max_amplitude;
  }


  Audio_Rec_Device::~Audio_Rec_Device()
  {
    close();
  }


  int Audio_Rec_Device::get_rec_rate()
  {
    return rec_rate;
  }


  bool Audio_Rec_Device::set_rec_rate(int rate)
  {
    ioctl(dev->audioDev, SNDCTL_DSP_SPEED, &rate);
    rec_rate = rate;
    return true;
  }


  int Audio_Rec_Device::get_rec_bits()
  {
    return rec_bits;
  }


  bool Audio_Rec_Device::set_rec_bits(int bits)
  {
    int fmt = bits==16 ? AFMT_S16_LE : AFMT_S8;

    if (bits!=8 && bits!=16)
      return false;
	
    ioctl(dev->audioDev, SNDCTL_DSP_SETFMT, &fmt);
    rec_bits = bits;
	
    return true;
  }


  bool Audio_Rec_Device::query_rec_stereo()
  {
    return rec_stereo;
  }


  bool Audio_Rec_Device::enable_rec_stereo(bool on)
  {
    ioctl(dev->audioDev, SNDCTL_DSP_STEREO, &on);
    rec_stereo = on;
    return true;
  }


  int Audio_Rec_Device::get_rec_buffersize()
  {
    int size;
    ioctl(dev->audioDev, SNDCTL_DSP_GETBLKSIZE, &size);
    return size;
  }


  // TODO: This should probably be implemented
  bool Audio_Rec_Device::set_rec_buffersize(int size)
  {
    return false;
  }

  bool Audio_Rec_Device::record(vec &mono, int n)
  {
    int i;

    if (rec_stereo) {
      bool ret;
      vec left(n), right(n);
	
      ret = record(left, right, n);
      mono = left + right / 2.0;
      return ret;
    }

    mono.set_size(n, false);
    if (rec_bits == 8) {
      signed char data[n];

      if (read(data, n) != n)
	return false;
      for (i=0; i<n; i++)
	mono(i) = (double)data[i] * rec_max_ampl / 127.0;
    } else {
      short data[n];

      if (read(data, n) != n)
	return false;
      for (i=0; i<n; i++)
	mono(i) = (double)data[i] * rec_max_ampl / 32767.0;
    }

    return true;
  }

  bool Audio_Rec_Device::record(vec &left, vec &right, int n)
  {
    int i;

    if (!rec_stereo) {
      bool ret;
	
      ret = record(left, n);
      right = left;
      return ret;
    }

    left.set_size(n, false);
    right.set_size(n, false);
    if (rec_bits == 8) {
      signed char data[2 * n];

      if (read(data, 2 * n) != 2 * n)
	return false;
      for (i=0; i<n; i++) {
	left(i) = (double)data[2*i] * rec_max_ampl / 127.0;
	right(i) = (double)data[2*i+1] * rec_max_ampl / 127.0;
      }
    } else {
      short data[2 * n];

      if (read(data, 2 * n) != 2 * n)
	return false;
      for (i=0; i<n; i++) {
	left(i) = (double)data[2*i] * rec_max_ampl / 32767.0;
	right(i) = (double)data[2*i+1] * rec_max_ampl / 32767.0;
      }
    }

    return true;
  }


  void Audio_Rec_Device::stop_record()
  {
    it_error("Audio_Record_Device::stop_record() needs to be written!");
  }


  //
  // Audio_IO_Device
  //
  Audio_IO_Device::Audio_IO_Device(double max_amplitude,
				   int rate, int bits, bool stereo)
    : Audio_Play_Device(max_amplitude),
      Audio_Rec_Device(max_amplitude)
  {
    it_assert1(open(mode_rdwr), "Cannot open audio device!");
	
    set_rate(rate);
    set_bits(bits);
    enable_stereo(stereo);
  }


  Audio_IO_Device::~Audio_IO_Device()
  {
    close();
  }


  bool Audio_IO_Device::set_rate(int rate)
  {
    return (set_play_rate(rate) && set_rec_rate(rate));
  }


  bool Audio_IO_Device::set_bits(int bits)
  {
    return (set_play_bits(bits) && set_rec_bits(bits));
  }


  bool Audio_IO_Device::enable_stereo(bool on)
  {
    return (enable_play_stereo(on) && enable_rec_stereo(on));
  }


  bool Audio_IO_Device::set_buffersize(int size)
  {
    return (set_play_buffersize(size) && set_rec_buffersize(size));
  }


  void Audio_IO_Device::stop()
  {
    it_error("Audio_IO_Device::stop() needs to be written!");
  }

} //namespace itpp

#endif
// ---------------- end Linux audio code ---------------------

// ---------------- Sun audio code ----------------------------------------
#ifdef sun

#include "base/operators.h"

#include <sys/audioio.h>

#include <iostream>
#include <cstdlib>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <stropts.h>
#include <sys/conf.h>

namespace itpp {

#define AUDIO_DEVICE "/dev/audio"

  inline static signed char clip_to_char(double x)
  {
    if (x < -127.0)
      return -127;
    else if (x > 127.0)
      return 127;
    else
      return (signed char)(x+0.5);
  }


  inline static short clip_to_short(double x)
  {
    if (x < -32767.0)
      return -32767;
    else if (x > 32767.0)
      return 32767;
    else
      return (short)(x+0.5);
  }

  struct Audio_Basic_Device::Device {
    int audioDev;
  };

  //
  // Audio_Basic_Device
  //
  Audio_Basic_Device::Audio_Basic_Device()
  {
    dev = new Device;
    isOpen = false;
  }

  Audio_Basic_Device::~Audio_Basic_Device()
  {
    close();
    delete dev;
  }

  bool Audio_Basic_Device::open(open_mode mode)
  {
    int flags = 0;
	
    switch (mode) {
    case mode_rd :
      flags = O_RDONLY;
      break;
    case mode_wr :
      flags = O_WRONLY;
      break;
    case mode_rdwr :
      flags = O_RDWR;
      break;
    }
	
    dev->audioDev = ::open(AUDIO_DEVICE, flags);
    isOpen = (dev->audioDev != -1);
	
    return isOpen;
  }

  void Audio_Basic_Device::close()
  {
    if (isOpen)
      ::close(dev->audioDev);
    isOpen = false;
  }

  int Audio_Basic_Device::read(signed char *p, int n)
  {
    return ::read(dev->audioDev, p, n * sizeof(signed char)) / sizeof(signed char);
  }

  int Audio_Basic_Device::read(short *p, int n)
  {
    return ::read(dev->audioDev, p, n * sizeof(short)) / sizeof(short);
  }

  int Audio_Basic_Device::write(const signed char *p, int n)
  {
    return ::write(dev->audioDev, p, n * sizeof(signed char)) / sizeof(signed char);
  }


  int Audio_Basic_Device::write(const short *p, int n)
  {
    return ::write(dev->audioDev, p, n * sizeof(short)) / sizeof(short);
  }

  //
  // Audio_Play_Device
  //
  Audio_Play_Device::Audio_Play_Device(double max_amplitude,
				       int rate, int bits, bool stereo)
  {
    struct audio_info ai;

    play_max_ampl = max_amplitude;
    it_assert1(open(mode_wr), "Cannot open audio device!");
    
    AUDIO_INITINFO(&ai);
    ai.play.encoding = AUDIO_ENCODING_LINEAR;
    ioctl(dev->audioDev, AUDIO_SETINFO, &ai);

    set_play_rate(rate);
    set_play_bits(bits);
    enable_play_stereo(stereo);
  }
	
  Audio_Play_Device::Audio_Play_Device(double max_amplitude)
  {
    play_max_ampl = max_amplitude;
  }

  Audio_Play_Device::~Audio_Play_Device()
  {
    close();
  }

  int Audio_Play_Device::get_play_rate()
  {
    return play_rate;
  }

  bool Audio_Play_Device::set_play_rate(int rate)
  {
    struct audio_info ai;
    
    AUDIO_INITINFO(&ai);
    ai.play.sample_rate = rate;
    ioctl(dev->audioDev, AUDIO_SETINFO, &ai);
    play_rate = ai.play.sample_rate;
    
    return play_rate == rate;
  }

  int Audio_Play_Device::get_play_bits()
  {
    return play_bits;
  }

  bool Audio_Play_Device::set_play_bits(int bits)
  {
    struct audio_info ai;

    AUDIO_INITINFO(&ai);
    ai.play.precision = bits;
    ioctl(dev->audioDev, AUDIO_SETINFO, &ai);
    play_bits = ai.play.precision;
	
    return play_bits == bits;
  }

  bool Audio_Play_Device::query_play_stereo()
  {
    return play_stereo;
  }

  bool Audio_Play_Device::enable_play_stereo(bool on)
  {
    struct audio_info ai;
    int channels = on ? 2 : 1;
    
    AUDIO_INITINFO(&ai);
    ai.play.channels = channels;
    ioctl(dev->audioDev, AUDIO_SETINFO, &ai);
    play_stereo = ai.play.channels==1 ? false : true;
    
    return play_stereo == on;
  }

  int Audio_Play_Device::get_play_buffersize()
  {
    struct audio_info ai;
    
    ioctl(dev->audioDev, AUDIO_GETINFO, &ai);
    
    return ai.play.buffer_size;
  }

  bool Audio_Play_Device::set_play_buffersize(int size)
  {
    struct audio_info ai;

    it_assert1(size >= 0, "Buffer size must be >= 0 in set_play_buffersize()");
    AUDIO_INITINFO(&ai);
    ai.play.buffer_size = (unsigned)size;
    ioctl(dev->audioDev, AUDIO_SETINFO, &ai);
    
    return ai.play.buffer_size == (unsigned)size;
  }

  bool Audio_Play_Device::play(const vec &mono)
  {
    int i, n;
  
    if (play_stereo)
      return play(mono, mono);
  
    if (play_bits == 8) {
      //signed char data[mono.size()];
      signed char *data = new signed char[mono.size()]; //CC-Fix
      for (i=0; i<mono.size(); i++)
	data[i] = clip_to_char(mono(i)*127.0/play_max_ampl);
      n = write(data, mono.size());
      delete[] data; //CC-fix
    } else {
      //short data[mono.size()];
      short *data = new short[mono.size()]; //CC-Fix
      for (i=0; i<mono.size(); i++)
	data[i] = clip_to_short(mono(i)*32767.0/play_max_ampl);
      n = write(data, mono.size());
      delete[] data; //CC-fix
    }
  
    return n == mono.size();
  }

  bool Audio_Play_Device::play(const vec &left, const vec &right)
  {
    int i, n;
  
    it_assert1(left.size() == right.size(), "Left and right channel must have "
	       "the same size in play()!");
  
    if (!play_stereo)
      return play(left + right / 2.0);
  
    if (play_bits == 8) {
      //signed char data[2 * left.size()];
      signed char *data = new signed char[2 * left.size()]; //CC-Fix    
      for (i=0; i<left.size(); i++) {
	data[2*i] = clip_to_char(left(i)*127.0/play_max_ampl);
	data[2*i+1] = clip_to_char(right(i)*127.0/play_max_ampl);
      }
      n = write(data, 2 * left.size()) / 2;
      delete[] data; //CC-fix
    } else {
      //short data[2 * left.size()];
      short *data = new short[2 * left.size()]; //CC-Fix
      for (i=0; i<left.size(); i++) {
	data[2*i] = clip_to_short(left(i)*32767.0/play_max_ampl);
	data[2*i+1] = clip_to_short(right(i)*32767.0/play_max_ampl);
      }
      n = write(data, 2 * left.size()) / 2;
      delete[] data; //CC-fix
    }
  
    return n == left.size();
  }

  void Audio_Play_Device::stop_play()
  {
    ioctl(dev->audioDev, I_FLUSH, FLUSHW);
  }

  void Audio_Play_Device::flush()
  {
    ioctl(dev->audioDev, AUDIO_DRAIN, 0);
  }

  //
  // Audio_Rec_Device
  //
  Audio_Rec_Device::Audio_Rec_Device(double max_amplitude,
				     int rate, int bits, bool stereo)
  {
    struct audio_info ai;
    
    rec_max_ampl = max_amplitude;
    
    it_assert1(open(mode_rd), "Cannot open audio device!");

    AUDIO_INITINFO(&ai);
    ai.record.encoding = AUDIO_ENCODING_LINEAR;
    ioctl(dev->audioDev, AUDIO_SETINFO, &ai);

    set_rec_rate(rate);
    set_rec_bits(bits);
    enable_rec_stereo(stereo);
  }

  Audio_Rec_Device::Audio_Rec_Device(double max_amplitude)
  {
    rec_max_ampl = max_amplitude;
  }

  Audio_Rec_Device::~Audio_Rec_Device()
  {
    close();
  }

  int Audio_Rec_Device::get_rec_rate()
  {
    return rec_rate;
  }

  bool Audio_Rec_Device::set_rec_rate(int rate)
  {
    struct audio_info ai;
    
    AUDIO_INITINFO(&ai);
    ai.record.sample_rate = rate;
    ioctl(dev->audioDev, AUDIO_SETINFO, &ai);
    rec_rate = ai.record.sample_rate;
    
    return rec_rate == rate;
  }

  int Audio_Rec_Device::get_rec_bits()
  {
    return rec_bits;
  }

  bool Audio_Rec_Device::set_rec_bits(int bits)
  {
    struct audio_info ai;

    AUDIO_INITINFO(&ai);
    ai.record.precision = bits;
    ioctl(dev->audioDev, AUDIO_SETINFO, &ai);
    rec_bits = ai.record.precision;
	
    return rec_bits == bits;
  }

  bool Audio_Rec_Device::query_rec_stereo()
  {
    return rec_stereo;
  }

  bool Audio_Rec_Device::enable_rec_stereo(bool on)
  {
    struct audio_info ai;
    int channels = on ? 2 : 1;
    
    AUDIO_INITINFO(&ai);
    ai.record.channels = channels;
    ioctl(dev->audioDev, AUDIO_SETINFO, &ai);
    rec_stereo = ai.record.channels==1 ? false : true;
    
    return rec_stereo == on;
  }

  int Audio_Rec_Device::get_rec_buffersize()
  {
    struct audio_info ai;
    
    ioctl(dev->audioDev, AUDIO_GETINFO, &ai);
    
    return ai.record.buffer_size;
  }

  bool Audio_Rec_Device::set_rec_buffersize(int size)
  {
    struct audio_info ai;
    
    it_assert1(size >= 0, "Buffer size must be >= 0 in set_play_buffersize()");
    AUDIO_INITINFO(&ai);
    ai.record.buffer_size = (unsigned)size;
    ioctl(dev->audioDev, AUDIO_SETINFO, &ai);
    
    return ai.record.buffer_size == (unsigned)size;
  }

  bool Audio_Rec_Device::record(vec &mono, int n)
  {
    int i;
  
    if (rec_stereo) {
      bool ret;
      vec left(n), right(n);
    
      ret = record(left, right, n);
      mono = left + right / 2.0;
      return ret;
    }
  
    mono.set_size(n, false);
    if (rec_bits == 8) {
      //signed char data[n];
      signed char *data = new signed char[n]; //CC-Fix

      if (read(data, n) != n)
	return false;
      for (i=0; i<n; i++) {
	mono(i) = (double)data[i] * rec_max_ampl / 127.0;
      }
      delete[] data; //CC-fix

    } else {
      //short data[n];
      short *data = new short[n]; //CC-Fix
    
      if (read(data, n) != n)
	return false;
      for (i=0; i<n; i++) {
	mono(i) = (double)data[i] * rec_max_ampl / 32767.0;
      }
      delete[] data; //CC-fix

    }
  
    return true;
  }

  bool Audio_Rec_Device::record(vec &left, vec &right, int n)
  {
    int i;
  
    if (!rec_stereo) {
      bool ret;
    
      ret = record(left, n);
      right = left;
      return ret;
    }
  
    left.set_size(n, false);
    right.set_size(n, false);
    if (rec_bits == 8) {
      //signed char data[2 * n];
      signed char *data = new signed char[2 * n]; //CC-Fix
    
      if (read(data, 2 * n) != 2 * n)
	return false;
      for (i=0; i<n; i++) {
	left(i) = (double)data[2*i] * rec_max_ampl / 127.0;
	right(i) = (double)data[2*i+1] * rec_max_ampl / 127.0;
      }
      delete[] data; //CC-fix

    } else {
      //short data[2 * n];
      short *data = new short[2 * n]; //CC-Fix
    
      if (read(data, 2 * n) != 2 * n)
	return false;
      for (i=0; i<n; i++) {
	left(i) = (double)data[2*i] * rec_max_ampl / 32767.0;
	right(i) = (double)data[2*i+1] * rec_max_ampl / 32767.0;
      }
      delete[] data; //CC-fix

    }
  
    return true;
  }

  void Audio_Rec_Device::stop_record()
  {
    ioctl(dev->audioDev, I_FLUSH, FLUSHR);
  }

  //
  // Audio_IO_Device
  //
  Audio_IO_Device::Audio_IO_Device(double max_amplitude,
				   int rate, int bits, bool stereo)
    : Audio_Play_Device(max_amplitude),
      Audio_Rec_Device(max_amplitude)
  {
    struct audio_info ai;
    
    it_assert1(open(mode_rdwr), "Cannot open audio device!");
	
    AUDIO_INITINFO(&ai);
    ai.play.encoding = AUDIO_ENCODING_LINEAR;
    ai.record.encoding = AUDIO_ENCODING_LINEAR;
    ioctl(dev->audioDev, AUDIO_SETINFO, &ai);

    set_rate(rate);
    set_bits(bits);
    enable_stereo(stereo);
  }

  Audio_IO_Device::~Audio_IO_Device()
  {
    close();
  }

  bool Audio_IO_Device::set_rate(int rate)
  {
    return (set_play_rate(rate) && set_rec_rate(rate));
  }

  bool Audio_IO_Device::set_bits(int bits)
  {
    return (set_play_bits(bits) && set_rec_bits(bits));
  }

  bool Audio_IO_Device::enable_stereo(bool on)
  {
    return (enable_play_stereo(on) && enable_rec_stereo(on));
  }

  bool Audio_IO_Device::set_buffersize(int size)
  {
    return (set_play_buffersize(size) && set_rec_buffersize(size));
  }

  void Audio_IO_Device::stop()
  {
    ioctl(dev->audioDev, I_FLUSH, FLUSHRW);
  }

} //namespace itpp

#endif
// ---------------- end Sun audio code ----------------------------------------

// ---------------- Windows audio code ----------------------------------------
#ifdef _MSC_VER

#include <windows.h>

namespace itpp {

  inline static signed char clip_to_char(double x)
  {
    if (x < -127.0)
      return -127;
    else if (x > 127.0)
      return 127;
    else
      return (signed char)(x+0.5);
  }

  inline static short clip_to_short(double x)
  {
    if (x < -32767.0)
      return -32767;
    else if (x > 32767.0)
      return 32767;
    else
      return (short)(x+0.5);
  }

  struct Audio_Basic_Device::Device {
    HWAVEOUT audioOutDev;
    WAVEFORMATEX audioOutFormat;
  };

  //
  // Audio_Basic_Device
  //
  Audio_Basic_Device::Audio_Basic_Device()
  {
    //    audioInFormat.wFormatTag = WAVE_FORMAT_PCM;
    //    audioInFormat.nChannels = 1;
    //    audioInFormat.nSamplesPerSec = 8000;
    //    audioInFormat.nAvgBytesPerSec = 16000;
    //    audioInFormat.nBlockAlign = 2;
    //    audioInFormat.wBitsPerSample = 16;
    //    audioInFormat.cbSize = 0;
    dev = new Device;
    dev->audioOutFormat.wFormatTag = WAVE_FORMAT_PCM;
    dev->audioOutFormat.nChannels = 1;
    dev->audioOutFormat.nSamplesPerSec = 8000;
    dev->audioOutFormat.nAvgBytesPerSec = 16000;
    dev->audioOutFormat.nBlockAlign = 2;
    dev->audioOutFormat.wBitsPerSample = 16;
    dev->audioOutFormat.cbSize = 0;
  
    isOpen = false;
  }

  Audio_Basic_Device::~Audio_Basic_Device()
  {
    close();
    delete dev;
  }

  bool Audio_Basic_Device::open(open_mode mode)
  {
    MMRESULT result2;

    //    audioInFormat.wFormatTag = WAVE_FORMAT_PCM;
    //    audioInFormat.nChannels = 1;
    //    audioInFormat.nSamplesPerSec = 8000;
    //    audioInFormat.nAvgBytesPerSec = 16000;
    //    audioInFormat.nBlockAlign = 2;
    //    audioInFormat.wBitsPerSample = 16;
    //    audioInFormat.cbSize = 0;
    dev->audioOutFormat.wFormatTag = WAVE_FORMAT_PCM;
    dev->audioOutFormat.nChannels = 1;
    dev->audioOutFormat.nSamplesPerSec = 8000;
    dev->audioOutFormat.nAvgBytesPerSec = 16000;
    dev->audioOutFormat.nBlockAlign = 2;
    dev->audioOutFormat.wBitsPerSample = 16;
    dev->audioOutFormat.cbSize = 0;

    switch (mode) {
    case mode_rd :
      //result1 = waveInOpen(&audioInDev, WAVE_MAPPER, &audioInFormat, 0, 0, CALLBACK_NULL);
      //isOpen = result1==MMSYSERR_NOERROR;
      break;
    case mode_wr :
      result2 = waveOutOpen(&dev->audioOutDev, WAVE_MAPPER, &dev->audioOutFormat, 0, 0, CALLBACK_NULL);
      isOpen = result2==MMSYSERR_NOERROR;
      break;
    case mode_rdwr :
      //result1 = waveInOpen(&audioInDev, WAVE_MAPPER, &audioInFormat, 0, 0, CALLBACK_NULL);
      result2 = waveOutOpen(&dev->audioOutDev, WAVE_MAPPER, &dev->audioOutFormat, 0, 0, CALLBACK_NULL);
      //isOpen = result1==MMSYSERR_NOERROR && result2==MMSYSERR_NOERROR;
      isOpen = result2==MMSYSERR_NOERROR;
      break;
    }
	
    return isOpen;
  }

  void Audio_Basic_Device::close()
  {
    //waveInClose(audioInDev);
    waveOutClose(dev->audioOutDev);
    isOpen = false;
  }

  int Audio_Basic_Device::read(signed char *p, int n)
  {
    it_error("Recording not supported");
    return 0;
  }

  int Audio_Basic_Device::read(short *p, int n)
  {
    it_error("Recording not supported");
    return 0;
  }

  int Audio_Basic_Device::write(const signed char *p, int n)
  {
    WAVEHDR hdr;
    MMRESULT status;

    hdr.lpData = (LPSTR)p;
    hdr.dwBufferLength = n;
    hdr.dwBytesRecorded = 0;
    hdr.dwUser = 0;
    hdr.dwFlags = 0;
    hdr.dwLoops = 1;

    status = waveOutPrepareHeader(dev->audioOutDev, &hdr, sizeof(hdr));
    if (status != MMSYSERR_NOERROR)
      return 0;
    status = waveOutWrite(dev->audioOutDev, &hdr, sizeof(hdr));
    if (status != MMSYSERR_NOERROR)
      return 0;

    while ((hdr.dwFlags & WHDR_DONE) == 0)
      ;
    
    return n;
  }

  int Audio_Basic_Device::write(const short *p, int n)
  {
    volatile WAVEHDR hdr;
    MMRESULT status;

    hdr.lpData = (LPSTR)p;
    hdr.dwBufferLength = n * sizeof(short);
    hdr.dwBytesRecorded = 0;
    hdr.dwUser = 0;
    hdr.dwFlags = 0;
    hdr.dwLoops = 1;

    status = waveOutPrepareHeader(dev->audioOutDev, (LPWAVEHDR)&hdr, sizeof(hdr));
    if (status != MMSYSERR_NOERROR)
      return 0;
    status = waveOutWrite(dev->audioOutDev, (LPWAVEHDR)&hdr, sizeof(hdr));
    if (status != MMSYSERR_NOERROR)
      return 0;

    while ((hdr.dwFlags & WHDR_DONE) == 0)
      ;
    
    return n;
  }

  //
  // Audio_Play_Device
  //
  Audio_Play_Device::Audio_Play_Device(double max_amplitude,
				       int rate, int bits, bool stereo)
  {
    play_max_ampl = max_amplitude;
    it_assert1(open(mode_wr), "Cannot open audio device!");
    
    set_play_rate(rate);
    set_play_bits(bits);
    enable_play_stereo(stereo);
  }
	
  Audio_Play_Device::Audio_Play_Device(double max_amplitude)
  {
    play_max_ampl = max_amplitude;
  }

  Audio_Play_Device::~Audio_Play_Device()
  {
    close();
  }

  int Audio_Play_Device::get_play_rate()
  {
    return play_rate;
  }

  bool Audio_Play_Device::set_play_rate(int rate)
  {
    // FIXME
    play_rate = rate;
    return true;
  }

  int Audio_Play_Device::get_play_bits()
  {
    return play_bits;
  }

  bool Audio_Play_Device::set_play_bits(int bits)
  {
    // FIXME
    play_bits = bits;
	
    return true;
  }

  bool Audio_Play_Device::query_play_stereo()
  {
    return play_stereo;
  }

  bool Audio_Play_Device::enable_play_stereo(bool on)
  {
    // FIXME
    play_stereo = on;
    return true;
  }

  int Audio_Play_Device::get_play_buffersize()
  {
    // FIXME
    return 0;
  }

  bool Audio_Play_Device::set_play_buffersize(int size)
  {
    // FIXME
    return false;
  }

  bool Audio_Play_Device::play(const vec &mono)
  {
    int i, n;

    if (play_stereo)
      return play(mono, mono);

    if (play_bits == 8) {
      signed char *ch_data = new signed char [mono.size()];
      for (i=0; i<mono.size(); i++)
	ch_data[i] = clip_to_char(mono(i)*127.0/play_max_ampl);
      n = write(ch_data, mono.size());
      delete [] ch_data;
    } else {
      short *s_data = new short[mono.size()];
      for (i=0; i<mono.size(); i++)
	s_data[i] = clip_to_short(mono(i)*32767.0/play_max_ampl);
      n = write(s_data, mono.size());
      delete [] s_data;
    }

    return n == mono.size();
  }

  bool Audio_Play_Device::play(const vec &left, const vec &right)
  {
    int i, n;

    it_assert1(left.size() == right.size(), "Left and right channel must have "
	       "the same size in play()!");
    
    if (!play_stereo)
      return play(left + right / 2.0);

    if (play_bits == 8) {
      signed char *ch_data = new signed char[2 * left.size()];
      for (i=0; i<left.size(); i++) {
	ch_data[2*i] = clip_to_char(left(i)*127.0/play_max_ampl);
	ch_data[2*i+1] = clip_to_char(right(i)*127.0/play_max_ampl);
      }
      n = write(ch_data, 2 * left.size()) / 2;
      delete [] ch_data;
    } else {
      short *s_data = new short[2 * left.size()];
      for (i=0; i<left.size(); i++) {
	s_data[2*i] = clip_to_short(left(i)*32767.0/play_max_ampl);
	s_data[2*i+1] = clip_to_short(right(i)*32767.0/play_max_ampl);
      }
      n = write(s_data, 2 * left.size()) / 2;
      delete [] s_data;
    }

    return n == left.size();
  }

  void Audio_Play_Device::stop_play()
  {
    it_error("Audio_Play_Device::stop_play() needs to be written!");
  }

  void Audio_Play_Device::flush()
  {
    it_error("Audio_Play_Device::flush() needs to be written!");
  }

  //
  // Audio_Rec_Device
  //
  Audio_Rec_Device::Audio_Rec_Device(double max_amplitude,
				     int rate, int bits, bool stereo)
  {
    rec_max_ampl = max_amplitude;
    
    it_assert1(open(mode_rd), "Cannot open audio device!");

    set_rec_rate(rate);
    set_rec_bits(bits);
    enable_rec_stereo(stereo);
  }

  Audio_Rec_Device::Audio_Rec_Device(double max_amplitude)
  {
    rec_max_ampl = max_amplitude;
  }

  Audio_Rec_Device::~Audio_Rec_Device()
  {
    close();
  }

  int Audio_Rec_Device::get_rec_rate()
  {
    return rec_rate;
  }

  bool Audio_Rec_Device::set_rec_rate(int rate)
  {
    // FIXME
    rec_rate = rate;
    return true;
  }

  int Audio_Rec_Device::get_rec_bits()
  {
    return rec_bits;
  }

  bool Audio_Rec_Device::set_rec_bits(int bits)
  {
    // FIXME
    rec_bits = bits;
	
    return true;
  }

  bool Audio_Rec_Device::query_rec_stereo()
  {
    return rec_stereo;
  }

  bool Audio_Rec_Device::enable_rec_stereo(bool on)
  {
    // FIXME
    rec_stereo = on;
    return true;
  }

  int Audio_Rec_Device::get_rec_buffersize()
  {
    // FIXME
    return 0;
  }

  bool Audio_Rec_Device::set_rec_buffersize(int size)
  {
    // FIXME
    return false;
  }

  bool Audio_Rec_Device::record(vec &mono, int n)
  {
    int i;

    if (rec_stereo) {
      bool ret;
      vec left(n), right(n);
	
      ret = record(left, right, n);
      mono = left + right / 2.0;
      return ret;
    }

    mono.set_size(n, false);
    if (rec_bits == 8) {
      signed char *ch_data = new signed char[n];
      if (read(ch_data, n) != n)
	return false;
      for (i=0; i<n; i++)
	mono(i) = (double)ch_data[i] * rec_max_ampl / 127.0;
      delete [] ch_data;
    } else {
      short *s_data = new short[n];
      if (read(s_data, n) != n)
	return false;
      for (i=0; i<n; i++)
	mono(i) = (double)s_data[i] * rec_max_ampl / 32767.0;
      delete [] s_data;
    }

    return true;
  }

  bool Audio_Rec_Device::record(vec &left, vec &right, int n)
  {
    int i;

    if (!rec_stereo) {
      bool ret;
	
      ret = record(left, n);
      right = left;
      return ret;
    }

    left.set_size(n, false);
    right.set_size(n, false);
    if (rec_bits == 8) {
      signed char *ch_data = new signed char[2 * n];
      if (read(ch_data, 2 * n) != 2 * n)
	return false;
      for (i=0; i<n; i++) {
	left(i) = (double)ch_data[2*i] * rec_max_ampl / 127.0;
	right(i) = (double)ch_data[2*i+1] * rec_max_ampl / 127.0;
      }
      delete [] ch_data;
    } else {
      short *s_data = new short[2 * n];
      if (read(s_data, 2 * n) != 2 * n)
	return false;
      for (i=0; i<n; i++) {
	left(i) = (double)s_data[2*i] * rec_max_ampl / 32767.0;
	right(i) = (double)s_data[2*i+1] * rec_max_ampl / 32767.0;
      }
      delete [] s_data;
    }

    return true;
  }

  void Audio_Rec_Device::stop_record()
  {
    it_error("Audio_Record_Device::stop_record() needs to be written!");
  }

  //
  // Audio_IO_Device
  //
  Audio_IO_Device::Audio_IO_Device(double max_amplitude,
				   int rate, int bits, bool stereo)
    : Audio_Play_Device(max_amplitude),
      Audio_Rec_Device(max_amplitude)
  {
    it_assert1(open(mode_rdwr), "Cannot open audio device!");
	
    set_rate(rate);
    set_bits(bits);
    enable_stereo(stereo);
  }

  Audio_IO_Device::~Audio_IO_Device()
  {
    close();
  }

  bool Audio_IO_Device::set_rate(int rate)
  {
    return (set_play_rate(rate) && set_rec_rate(rate));
  }

  bool Audio_IO_Device::set_bits(int bits)
  {
    return (set_play_bits(bits) && set_rec_bits(bits));
  }

  bool Audio_IO_Device::enable_stereo(bool on)
  {
    return (enable_play_stereo(on) && enable_rec_stereo(on));
  }

  bool Audio_IO_Device::set_buffersize(int size)
  {
    return (set_play_buffersize(size) && set_rec_buffersize(size));
  }

  void Audio_IO_Device::stop()
  {
    it_error("Audio_IO_Device::stop() needs to be written!");
  }

} //namespace itpp

#endif
// ---------------- Windows audio code ----------------------------------------
